# Copyright: 2007-2009 Paul Obermeier (obermeier@poSoft.de)
# Distributed under BSD license.

package provide tcomexcel 0.2

namespace eval ::Excel {

    # Only available in Excel 2003 and up.
    # TODO: Use these values instead of the mapFloatPoint variable.
    # $appId DecimalSeparator "."
    # $appId ThousandsSeparator ","
    # $appId UseSystemSeparators 0

    # The default settings for English Excel applications.
    variable language      "English" 
    variable mapFloatPoint 0

    # The default settings for German Excel applications.
    # Note, that if mapFloatPoint is set, ALL cell contents (independent of
    # the cell type) have the "," mapped to a "." when reading, and vice versa
    # when writing.
    # variable language      "German" 
    # variable mapFloatPoint 1

    variable errorValue    "UNKNOWN"

    proc GetErrorValue {} {
        variable errorValue

        return $errorValue
    }

    proc SetErrorValue { val } {
        variable errorValue

        set errorValue $val
    }

    proc SetMapFloatSeparator { onOff } {
        variable mapFloatPoint

        set mapFloatPoint $onOff
    }

    proc GetFloatSeparator {} {
        variable mapFloatPoint

        if { $mapFloatPoint } {
           return ","
        } else {
           return "."
        }
    }

    proc SetLanguage { lang } {
        variable language

        set language $lang
    }

    proc GetLangFuncName { funcName { addStr "" } } {
        variable language

        array set map {
            "MIN"     "MIN"
            "MAX"     "MAX"
            "AVERAGE" "MITTELWERT" 
            "STDEV"   "STABWN" 
            "TODAY"   "HEUTE" 
        }

        if { [lsearch [array names map] $funcName] < 0 } {
            error "Function name $funcName not in map"
        }

        switch $language {
            "German" {
                set langFuncName $map($funcName)
            }
            default {
                set langFuncName $funcName
            }
        }
        if { $addStr ne "" } {
            append langFuncName $addStr
        }
        return $langFuncName
    }

    proc GetLangNumberFormat { pre post } {
        set floatSep [::Excel::GetFloatSeparator]
        return [format "%s%s%s" $pre $floatSep $post]
    }

    proc TclBoolToExcel { val } {
        set boolInt 0
        if { $val } {
            set boolInt 1
        }
        return $boolInt
    }

    proc RgbToColor { r g b } {
        return [expr {$b << 16 | $g << 8 | $r}]
    }

    # Convert an Excel column string to a column number.
    proc ColumnCharToInt { colChar } {
        set abc {- A B C D E F G H I J K L M N O P Q R S T U V W X Y Z}
        set int 0
        foreach char [split $colChar ""] {
            set int [expr {$int*26 + [lsearch $abc $char]}]
        }
        #incr int -1 ; # one-letter columns start from A
        return $int
    }

    # Convert a column number to an Excel column string.
    proc ColumnIntToChar { colNum } {
        if { $colNum <= 26 } {
            return [format "%c" [expr {$colNum+64}]]
        } elseif { $colNum <= 676 } {
            set c1 [expr {int($colNum/26)+64}]
            set c2 [expr {$colNum%26+64}]
            return [format "%c%c" $c1 $c2]
        } else {
            set c1 [expr {int($colNum/676)+64}]
            set c2 [expr {int(($colNum-676)/26)+64}]
            set c3 [expr {$colNum%26+64}]
            return [format "%c%c%c" $c1 $c2 $c3]
        }
    } 

    # Convert a numeric cell range to an Excel range string.
    # The numeric range is specified as upper left and lower right corner.
    proc GetCellRange { rowNum1 colNum1 rowNum2 colNum2 } {
        set range [format "%s%d:%s%d" \
                   [ColumnIntToChar $colNum1] $rowNum1 \
                   [ColumnIntToChar $colNum2] $rowNum2]  
        return $range
    }

    proc GetColumnRange { colNum1 colNum2 } {
        set range [format "%s:%s" \
                   [ColumnIntToChar $colNum1] \
                   [ColumnIntToChar $colNum2]]  
        return $range
    }

    # Get the number of rows of a range.
    # "rangeId" can be an object of type range, cells, worksheet.
    # If specifying a worksheetId or cellsId, the maximum number of rows
    # of a worksheet will be returned.
    proc GetNumRows { rangeId } {
        set range [$rangeId Rows]
        return [$range Count]
    }

    # Get the number of rows of a worksheet in use.
    proc GetNumUsedRows { worksheetId } {
        set range [$worksheetId UsedRange]
        return [[$range Rows] Count]
    }

    # Get the number of columns of a range.
    # "rangeId" can be an object of type range, cells, worksheet.
    # If specifying a worksheetId or cellsId, the maximum number of columns
    # of a worksheet will be returned.
    proc GetNumColumns { rangeId } {
        set range [$rangeId Columns]
        return [$range Count]
    }

    # Get the number of columns of a worksheet in use.
    # In some cases the number of columns returned may be 1 to high.
    proc GetNumUsedColumns { worksheetId } {
        set range [$worksheetId UsedRange]
        return [[$range Columns] Count]
    }

    # Select a range by specifying an Excel range string.
    # The range is selected in worksheet/cells specified by cellsId.
    # Set visSel to true, to see the selection in the user interface.
    proc SelectRangeByStr { cellsId rangeStr { visSel false } } {
        set rangeId [$cellsId Range $rangeStr]
        if { $visSel } {
            $rangeId Select
        }
        return $rangeId
    }

    # Select a range by specifying a numeric range.
    # The range is selected in worksheet/cells specified by cellsId.
    # The numeric range is specified as upper left and lower right corner.
    # Set visSel to true, to see the selection in the user interface.
    proc SelectRangeByIndex { cellsId rowNum1 colNum1 rowNum2 colNum2 \
                              { visSel false } } {
        set rangeStr [GetCellRange $rowNum1 $colNum1 $rowNum2 $colNum2]
        set rangeId [$cellsId Range $rangeStr]
        if { $visSel } {
            $rangeId Select
        }
        return $rangeId
    }

    proc SelectAll { cellsId } {
        return [[GetApplicationId $cellsId] Cells]
    }

    proc SetRangeFontBold { rangeId { onOff true } } {
        set fontId [$rangeId Font]
        $fontId Bold [TclBoolToExcel $onOff]
    }

    # Set the horizontal alignment of a cell range.
    # "align" must be a value of enumeration XlHAlign (see excelConst.tcl).
    proc SetRangeHorizontalAlignment { rangeId align } {
        $rangeId HorizontalAlignment [expr $align]
    }

    # Set the vertical alignment of a cell range.
    # "align" must be a value of enumeration XlVAlign (see excelConst.tcl).
    proc SetRangeVerticalAlignment { rangeId align } {
        $rangeId VerticalAlignment [expr $align]
    }

    # Toggle the AutoFilter switch of a cell range.
    proc ToggleAutoFilter { rangeId } {
        $rangeId AutoFilter
    }

    # Set the fill color of a cell range.
    # The red, green and blue components must be integers from 0 to 255.
    proc SetRangeFillColor { rangeId r g b } {
        set colorNum [RgbToColor [expr {int($r)}] [expr {int($g)}] [expr {int($b)}]]
        [$rangeId Interior] Color $colorNum
        [$rangeId Interior] Pattern $::Excel::xlSolid
    }

    # Set the text color of a cell range.
    # The red, green and blue components must be integers from 0 to 255.
    proc SetRangeTextColor { rangeId r g b } {
        set colorNum [RgbToColor [expr {int($r)}] [expr {int($g)}] [expr {int($b)}]]
        [$rangeId Font] Color $colorNum
    }

    # Get the version of an Excel application.
    # By default the version is returned as version number (ex. 9.0).
    # If useString is set to true, the version name is returned (ex. Excel 2000). 
    proc GetVersion { appId { useString false } } {
        array set map {
            "8.0"  "Excel 97"
            "9.0"  "Excel 2000"
            "10.0" "Excel 2002"
            "11.0" "Excel 2003"
            "12.0" "Excel 2007"
        }
        set version [$appId Version]
        if { $useString } {
            if { [info exists map($version)] } {
                return $map($version) 
            } else {
                return "Unknown Excel version"
            }
        } else {
            return $version
        }
    }

    # Return the default extension of an Excel file.
    # Starting with Excel 12 (2007) this is the string ".xlsx".
    # In previous versions it was ".xls".
    proc GetExtString { appId } {
        switch [GetVersion $appId] {
            "12.0"  { return ".xlsx" }
            default { return ".xls" }
        }
    }

    # Return the active printer as a string.
    proc GetActivePrinter { appId } {
        return [$appId ActivePrinter]
    }

    # Return the current Excel user as a string.
    proc GetUserName { appId } {
        return [$appId UserName]
    }

    # Return the startup pathname.
    proc GetStartupPath { appId } {
        return [$appId StartupPath]
    }

    # Return the templates pathname.
    proc GetTemplatesPath { appId } {
        return [$appId TemplatesPath]
    }

    # Return the add-ins pathname.
    proc GetUserLibraryPath { appId } {
        return [$appId UserLibraryPath]
    }

    # Return the installation pathname.
    proc GetInstallationPath { appId } {
        return [$appId Path]
    }

    # Return the user folder's pathname.
    proc GetUserPath { appId } {
        return [$appId DefaultFilePath]
    }

    proc ShowAlerts { appId onOff } {
        $appId DisplayAlerts [TclBoolToExcel $onOff]
    }

    proc GetApplicationId { componentId } {
        return [$componentId Application]
    }

    proc GetApplicationName { applId } {
        return [$applId Name]
    }

    # Open a new Excel application.
    # If "visible" is set to false, the application starts in detached mode.
    # Return the identifier for the Excel application.
    proc OpenNew { { visible true } { width -1 } { height -1 } } {
        set retVal [catch {::tcom::ref createobject "Excel.Application"} appId]
        if { $retVal == 0 } {
            $appId Visible [TclBoolToExcel $visible]
        } else {
            error "Cannot start Excel application"
        }
	if { $width >= 0 } {
            $appId Width [expr $width]
        }
        if { $height >= 0 } {
            $appId Height [expr $height]
        }
        return $appId
    }

    # Open an Excel application. Use an already open Excel, if available.
    # If "visible" is set to false, the application starts in detached mode.
    # Return the identifier for the Excel application.
    proc Open { { visible true } { width -1 } { height -1 } } {
        set retVal [catch {::tcom::ref getactiveobject "Excel.Application"} appId]
        if { $retVal != 0 } {
            set retVal [catch {::tcom::ref createobject "Excel.Application"} appId]
            if { $retVal != 0 } {
                error "Cannot start Excel application"
            }
        }
        $appId Visible [TclBoolToExcel $visible]
        if { $width >= 0 } {
            $appId Width [expr $width]
        }
        if { $height >= 0 } {
            $appId Height [expr $height]
        }
        return $appId
    }

    # Quit the Excel application identified by "appId".
    # If saveChanges is set to false, the application quits without
    # saving any changes.
    proc Quit { appId { saveChanges true } } {
        if { ! $saveChanges } {
            ShowAlerts $appId false
        }
        $appId Quit
    }

    # Close a workbook without saving possible changes.
    # Use the SaveAs method before closing, if you want to save changes.
    proc Close { workbookId } {
        $workbookId Close [TclBoolToExcel false]
    }

    # Save a workbook in file "fileName".
    # If no "fmt" is given, the file is stored in native Excel format.
    # "fmt" must be a constant as defined by the xlFileFormat enumeration.
    # If "backup" is set to 1, a backup file is created before storing.
    proc SaveAs { workbookId fileName { fmt "" } { backup 0 } } {
        set fileName [file nativename $fileName]
        if { $fmt eq "" } {
            $workbookId SaveAs $fileName
        } else {
            set appId [GetApplicationId $workbookId]
            ShowAlerts $appId false
            $workbookId -namedarg SaveAs \
                        Filename $fileName \
                        FileFormat $fmt \
                        CreateBackup $backup
            ShowAlerts $appId true
        }
    }

    # Save a worksheet in CSV format in file "fileName".
    proc SaveAsCsv { workbookId worksheetId fileName } {
        set fileName [file nativename $fileName]
        set appId [GetApplicationId $workbookId]
        ShowAlerts $appId false
        $worksheetId -namedarg SaveAs \
                     Filename $fileName \
                     FileFormat $::Excel::xlCSV \
                     CreateBackup 0
        ShowAlerts $appId true
    }

    # This function adds a new workbook with 1 worksheet.
    # Valid enums for type are: 
    # xlChart
    # xlExcel4IntlMacroSheet
    # xlExcel4MacroSheet
    # xlWorksheet           (This is the default value)
    # Return the workbook identifier.
    proc AddWorkbook { appId { type "" } } {
        if { $type eq "" } {
            set type $::Excel::xlWorksheet
        }
        set workbooks [$appId Workbooks]
        set workbookId [$workbooks Add [expr $type]]
        return $workbookId
    }

    proc OpenWorkbook { appId fileName { readOnly false } } {
        set readOnlyInt [TclBoolToExcel $readOnly]
        set nativeName  [file nativename $fileName]
        set workbooks [$appId Workbooks]
        set retVal [catch {[$workbooks Item [file tail $fileName]] Activate} d]
        if { $retVal == 0 } {
            puts "$nativeName already open"
            set workbookId [$workbooks Item [file tail $fileName]]
        } else {
            set workbookId [$workbooks -namedarg Open Filename $nativeName \
                                                      ReadOnly $readOnlyInt]
        }
        return $workbookId
    }

    # Add a new worksheet to workbook "workbookId" with name "name". 
    # If visibleType is not given, the worksheet is visible.
    # Possible values for visibleType are:
    # xlSheetVisible, xlSheetHidden, xlSheetVeryHidden
    # Return the worksheet identifier.
    proc AddWorksheet { workbookId name { visibleType "" } } {
        set worksheets [$workbookId Worksheets]
        set lastWorksheet [$worksheets Item [$worksheets Count]]
        # The next line will insert the worksheet at the end.
        # This gives problems with Excel 2000, when adding charts.
        # set worksheetId [$worksheets Add [::tcom::na] $lastWorksheet]
        # So we add worksheets always at the beginning.
        set worksheetId [$worksheets Add]
        $worksheetId Name $name
        if { $visibleType ne "" } {
            $worksheetId Visible $visibleType
        }
        return $worksheetId
    }

    proc DeleteWorksheet { workbookId worksheetId } {
        set worksheets [$workbookId Worksheets]
        set count [$worksheets Count]

        if { $count == 1 } {
           return
        }

        # Delete the specified worksheet.
        # This will cause alert dialogs to be displayed unless they are turned off.
        set appId [GetApplicationId $workbookId]
        ShowAlerts $appId false
        $worksheetId Delete
        # Turn the alerts back on.
        ShowAlerts $appId true
    }

    proc DeleteWorksheetByIndex { workbookId index } {
        set worksheets [$workbookId Worksheets]
        set count [$worksheets Count]

        if { $count == 1 } {
           return
        }
        if { $index < 1 || $index > $count } {
            error "Invalid index $index given"
        }
        # Delete the specified worksheet.
        # This will cause alert dialogs to be displayed unless they are turned off.
        set appId [GetApplicationId $workbookId]
        ShowAlerts $appId false
        set worksheet [$worksheets Item $index]
        $worksheet Delete
        # Turn the alerts back on.
        ShowAlerts $appId true
    }

    # Return the worksheet identifier of worksheet "itemNum" of
    # workbook "workbookId".
    proc GetWorksheetIdByIndex { workbookId itemNum } {
        set worksheets [$workbookId Worksheets]
        set worksheetId [$worksheets Item [expr $itemNum]]
        return $worksheetId
    }

    # Return the worksheet identifier of worksheet with name "worksheetName".
    proc GetWorksheetIdByName { workbookId worksheetName } {
        set worksheets [$workbookId Worksheets]
        set count [$worksheets Count]
        for { set i 1 } { $i <= $count } { incr i } {
            set worksheetId [$worksheets Item [expr $i]]
            if { $worksheetName eq [$worksheetId Name] } {
                return $worksheetId
            }
        }
        error "No worksheet with name $worksheetName"
    }

    # Set the worksheet name to "name".
    proc SetWorksheetName { worksheetId name } {
        $worksheetId Name $name
    }

    # Return the worksheet name of worksheet identified by "worksheetId".
    proc GetWorksheetName { worksheetId } {
        set name [$worksheetId Name]
        return $name
    }

    # Return the number of worksheets in a workbook identified by "workbookId".
    proc GetNumWorksheets { workbookId } {
        set worksheets [$workbookId Worksheets]
        set count [$worksheets Count]
        return $count
    }

    # Return the cells identifier of worksheet identified by "worksheetId".
    proc GetCellsId { worksheetId } {
        set cellsId [$worksheetId Cells]
        return $cellsId
    }

    # Return the cell identifier of the cell indexed by (rowNum,ColNum).
    # The cell is contained in cells object identified by "cellsId".
    proc GetCellIdByIndex { cellsId rowNum colNum } {
        set colChar [ColumnIntToChar $colNum]
        set cell [$cellsId Item $rowNum $colNum]
        return $cell
    }

    # Select a cell by specifying a numeric range.
    # The range is selected in worksheet/cells specified by cellsId.
    # The numeric range is specified as row and column number.
    # Set visSel to true, to see the selection in the user interface.
    proc SelectCellByIndex { cellsId rowNum colNum { visSel false } } {
        set rangeStr [GetCellRange $rowNum $colNum $rowNum $colNum]
        set rangeId [$cellsId Range $rangeStr]
        if { $visSel } {
            $rangeId Select
        }
        return $rangeId
    }

    proc SetHyperlink { worksheetId rowNum colNum link { textDisplay "" } } {
        if { $textDisplay eq "" } {
            set textDisplay $link
        }

        set cellsId [GetCellsId $worksheetId]
        set rangeId [SelectRangeByIndex $cellsId $rowNum $colNum $rowNum $colNum]
        set hyperId [$worksheetId Hyperlinks]
        $hyperId -namedarg Add Anchor $rangeId \
                               Address $link \
                               TextToDisplay $textDisplay
    }

    proc InsertImage { worksheetId imgFileName { rowNum 0 } { colNum 0 } } {
        set cellId [SelectCellByIndex $worksheetId $rowNum $colNum true]
        set pictures [$worksheetId Pictures]
        set fileName [file nativename $imgFileName]
        set picId [$pictures Insert $imgFileName]
        return $picId
    }

    proc ScaleImage { picId scaleWidth scaleHeight } {
        set rangeId [$picId ShapeRange]
        $rangeId ScaleWidth  [expr $scaleWidth]  [TclBoolToExcel true]
        $rangeId ScaleHeight [expr $scaleHeight] [TclBoolToExcel true]
    }

    # Set the value of cell indexed by (rowNum,ColNum) to "val".
    # The cell is contained in cells object identified by "cellsId".
    proc SetCellValue { cellsId rowNum colNum val } {
        variable mapFloatPoint

        if { $mapFloatPoint } {
            set val [string map {"." ","} $val]
        }
        set colChar [ColumnIntToChar $colNum]
        $cellsId Item $rowNum $colNum $val
    }

    proc GetCellValue { cellsId rowNum colNum } {
        variable mapFloatPoint
        variable errorValue

        set colChar [ColumnIntToChar $colNum]
        set cell [$cellsId Item $rowNum $colNum]
        set retVal [catch {$cell Value} val]
        if { $retVal != 0 } {
            set val $errorValue
        }
        if { $mapFloatPoint } {
            set val [string map {"," "."} $val]
        }
        return $val
    }

    # Insert the values contained in valList into row rowNum of the
    # worksheet cells identified by cellsId.
    # The insertion is started in column startColNum.
    # By default all the list values are inserted. If the optional
    # parameter numVals is specified and greater than zero, then only
    # numVals are filled with the list values starting at list index 0.
    proc SetRowValues { cellsId rowNum startColNum valList { numVals 0 } } {
        set len [llength $valList]
        if { $numVals > 0 } {
            if { $numVals < $len } {
                set len $numVals
            }
        }
        set ind 0
        for { set c $startColNum } { $c <= [expr {$startColNum + $len}] } { incr c } {
            SetCellValue $cellsId $rowNum $c [lindex $valList $ind]
            incr ind
        }
    }

    # The fast way to do it. 
    # Does not work with Excel2003, but works with Excel2000.
    proc SetRowValuesFast { cellsId rowNum startColNum valList { numVals 0 } } {
        variable mapFloatPoint

        set len [llength $valList]
        if { $numVals > 0 } {
            if { $numVals < $len } {
                set len $numVals
            }
        }
        set rangeId [::Excel::SelectRangeByIndex $cellsId \
                     $rowNum $startColNum \
                     $rowNum [expr {$startColNum + $len -1}]]
        if { $mapFloatPoint } {
            # OPA TODO set valList [string map {"." ","} $valList]
        }
        $rangeId Value $valList
    }

    proc GetRowValues { cellsId rowNum startColNum { numVals 0 } } {
        if { $numVals <= 0 } {
            set len [GetNumColumns $cellsId]
        } else {
            set len $numVals
        }
        set valList [list]
        set colNum $startColNum
        set ind 0
        while { $ind < $len } {
            set val [GetCellValue $cellsId $rowNum $colNum]
            if { $val eq "" } {
                set val2 [GetCellValue $cellsId $rowNum [expr {$colNum+1}]]
                if { $val2 eq "" } {
                    break
                }
            }
            lappend valList $val
            incr ind
            incr colNum
        }
        return $valList
    }

    # Set the width of a column. 
    # A positive value specifies the column's width in average-size characters
    # of the widget's font. A value of zero specifies that the column's width
    # fits automatically the width of all elements in the column.
    proc SetColumnWidth { cellsId colNum width } {
        set cell [SelectRangeByIndex $cellsId 1 $colNum 1 $colNum]
        set curCol [$cell EntireColumn]
        if { $width == 0 } {
            [$curCol Columns] AutoFit
        } else {
            $curCol ColumnWidth $width
        }
    }

    proc SetColumnsWidth { cellsId startColNum endColNum width } {
        for { set c $startColNum } { $c <= $endColNum } { incr c } {
            SetColumnWidth $cellsId $c $width
        }
    }

    proc SetColumnValues { cellsId colNum startRowNum valList { numVals 0 } } {
        set len [llength $valList]
        if { $numVals > 0 } {
            if { $numVals < $len } {
                set len $numVals
            }
        }
        set ind 0
        for { set r $startRowNum } { $r <= [expr {$startRowNum + $len}] } { incr r } {
            SetCellValue $cellsId $r $colNum [lindex $valList $ind]
            incr ind
        }
    }

    proc GetColumnValues { cellsId colNum startRowNum { numVals 0 } } {
        if { $numVals <= 0 } {
            set len [GetNumRows $cellsId]
        } else {
            set len $numVals 
        }
        set valList [list]
        set rowNum $startRowNum
        set ind 0
        while { $ind < $len } {
            set val [GetCellValue $cellsId $rowNum $colNum]
            if { $val eq "" } {
                set val2 [GetCellValue $cellsId [expr {$rowNum+1}] $colNum]
                if { $val2 eq "" } {
                    break
                }
            }
            lappend valList $val
            incr ind
            incr rowNum
        }
        return $valList
    }
}
